package it.bigatti.iSafe;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.xml.parsers.*;
import javax.swing.*;
import javax.swing.event.*;

import org.w3c.dom.*;
import org.xml.sax.*;

import gokul.java.io.base64.*;
import com.strangeberry.rendezvous.*;

class BrowseFrame implements ServiceListener {
    
    JFrame frame;
    JPanel panel;
    JLabel status;
    
    DefaultListModel hostModel;
    DefaultListModel storageModel;
    
    AbstractAction addAction;
    AbstractAction refreshAction;
    
    java.util.List hosts = new LinkedList();
    Map hostsInfo = new HashMap();
    
    ServiceInfo currentService = null;
    
    public BrowseFrame() {
        createModels();
        createUI();
    }
    
    void createModels() {
        hostModel = new DefaultListModel();
        storageModel = new DefaultListModel();
        hostsInfo = new HashMap();
    }
    
    void createUI() {
        try {
            UIManager.setLookAndFeel( UIManager.getSystemLookAndFeelClassName() );
        } catch( ClassNotFoundException ex ) {
        } catch( InstantiationException ex ) {
        } catch( IllegalAccessException ex ) {
        } catch( UnsupportedLookAndFeelException ex ) {
        }
        
        panel = new JPanel( new BorderLayout() );

        final JList hostList = new JList( hostModel );
        hostList.addListSelectionListener( new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent e) {
                if( !e.getValueIsAdjusting() ) {
                    int index = hostList.getSelectedIndex();
                    serviceSelected( index );
                }
            }
        });

        JList storageList = new JList( storageModel );
        status = new JLabel("Status bar");
        
        JSplitPane splitPane = new JSplitPane( JSplitPane.HORIZONTAL_SPLIT );
        splitPane.setLeftComponent( new JScrollPane( hostList ) );
        splitPane.setRightComponent( new JScrollPane( storageList ) );
        
        JToolBar toolBar = new JToolBar();
        addAction = new AbstractAction( "Add" ) {
            public void actionPerformed( ActionEvent e ) {
                add();
            }
        };
        addAction.setEnabled( false );
        toolBar.add( addAction );
        
        refreshAction = new AbstractAction( "Refresh" ) {
            public void actionPerformed( ActionEvent e ) {
                refresh();
            }
        };
        refreshAction.setEnabled( false );
        toolBar.add( refreshAction );
        
        panel.add( toolBar, BorderLayout.NORTH );
        panel.add( splitPane, BorderLayout.CENTER );
        panel.add( status, BorderLayout.SOUTH );
        
        frame = new JFrame("iSafe servers");
        frame.setDefaultCloseOperation( WindowConstants.EXIT_ON_CLOSE );
        frame.getContentPane().add( panel );
        frame.pack();
        frame.setLocation( 10, 10 );
        frame.setVisible( true );
    }
    
    void setStatus( String text ) {
        status.setText( text );
    }
    
    void raiseError( Exception ex ) {
        ex.printStackTrace();
        raiseError( ex.getMessage() );
    }
    
    void raiseError( String message ) {
        JOptionPane.showMessageDialog( frame, message, "iSafe", JOptionPane.ERROR_MESSAGE );
    }
    
    void serviceSelected( int index ) {
        if( index >=0 && index <= hosts.size() ) {
            currentService = (ServiceInfo)hostsInfo.get( hosts.get( index ) );
            setStatus( "iSafe Server version = " + currentService.getPropertyString( "version" ) + 
                " address = " + currentService.getAddress() + " host = " + hosts.get( index ) );
            
            addAction.setEnabled( true );
            refreshAction.setEnabled( true );
            fetchStoredData( currentService );
        } else {
            storageModel.clear();
            
            addAction.setEnabled( false );
            refreshAction.setEnabled( false );
            
            panel.revalidate();
        }
    }
    
    void fetchStoredData( ServiceInfo info ) {
        String xml = "<?xml version=\"1.0\"?>"+
            "<SOAP-ENV:Envelope xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\">"+
            "  <SOAP-ENV:Body>"+
            "    <getStoredData/>"+
            "  </SOAP-ENV:Body>"+
            "</SOAP-ENV:Envelope>\n";
        
        storageModel.clear();
        
        try {
            Document doc = sendRequest( info, xml );
            if( doc != null ) {
                NodeList list = doc.getElementsByTagName("file");
                
                for( int i=0; i<list.getLength(); i++ ) {
                    Node node = list.item(i);
                    NamedNodeMap attributes = node.getAttributes();
                    
                    String element = attributes.getNamedItem( "name" ).getNodeValue() + " [" +
                        attributes.getNamedItem( "date" ).getNodeValue() + ", " +
                        attributes.getNamedItem( "size" ).getNodeValue() + "]";
        
                    storageModel.addElement( element );
                }
            }
        } catch( Exception ex ) {
            raiseError( ex );
        }
    }
    
    void add() {
        JFileChooser chooser = new JFileChooser();
        int returnVal = chooser.showOpenDialog( frame );
        if(returnVal == JFileChooser.APPROVE_OPTION) {
            
            try {
                File file = chooser.getSelectedFile();
                
                FileInputStream fis = new FileInputStream( file );
                ByteArrayOutputStream bas = new ByteArrayOutputStream();
                Base64OutputStream out = new Base64OutputStream( bas, true );
                
                byte[] buffer = new byte[ 4096 ];
                
                while( true ) {
                    int len = fis.read( buffer );
                    
                    if( len == -1 ) {
                        break;
                    }
                    
                    out.write( buffer, 0, len );
                }
                
                out.flush();
                out.close();
                fis.close();
                
                addFile( currentService, chooser.getSelectedFile().getName(), bas.toByteArray() );
            } catch( Exception ex ) {
                raiseError( ex );
            }
        }
    }
    
    void refresh() {
        fetchStoredData( currentService );
    }
    
    void addFile( ServiceInfo info, String filename, byte[] buffer ) throws IOException,
            SAXException, ParserConfigurationException {
        String xml1 = "<?xml version=\"1.0\"?>"+
            "<SOAP-ENV:Envelope xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\">"+
            "  <SOAP-ENV:Body>"+
            "    <addFile>";
            
        String xml3 = "    </addFile>"+
            "  </SOAP-ENV:Body>"+
            "</SOAP-ENV:Envelope>\n";
            
        String xml2 = "      <file name=\"" + filename + "\">" + new String( buffer ) + 
            "      </file>";
            
        Document doc = sendRequest( info, xml1 + xml2 + xml3 );
        refresh();
    }
    
    public void addService(Rendezvous rendezvous, String type, String name) {
        System.out.println("ADD: " + name);
        rendezvous.requestServiceInfo(type, name);
        
        if( !hosts.contains( name ) ) {
            hosts.add( name );
            hostModel.addElement( name );
            //System.out.println( hosts );
            panel.revalidate();
        }
    }

    public void resolveService(Rendezvous rendezvous, String type, String name, ServiceInfo info) {
        System.out.println("RESOLVE: " + info);
        hostsInfo.put( name, info );
    }

    public void removeService(Rendezvous rendezvous, String type, String name) {
        System.out.println("REMOVE: " + name);
        hosts.remove( name );
        hostModel.removeElement( name );
    }
    
    Document sendRequest( ServiceInfo info, String xml ) throws IOException, SAXException, 
            IllegalArgumentException, ParserConfigurationException {
                
        String address = info.getAddress();
        int port = info.getPort();
        Document response = null;
        
        //System.out.println("BrowseFrame.sendRequest(): address="+address+" port="+port);
        
        Socket socket = new Socket( address, port );
        DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
        BufferedReader in = new BufferedReader( new InputStreamReader( socket.getInputStream() ) );

        out.writeBytes( Service.PREFIX );
        out.writeBytes( "" + xml.length() );
        out.writeBytes( "\n" );
        out.writeBytes( xml );
        
        System.out.println( "C->S: " + xml );

        String responseXml = in.readLine();
        //System.out.println( "IN: " + responseXml );
        
        if( responseXml != null && responseXml.length()>0 ) {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            response = builder.parse( new ByteArrayInputStream( responseXml.getBytes( ) ) );
            
            NodeList list = response.getElementsByTagName( "Fault");
            if( list.getLength() > 0 ) {
                list = response.getElementsByTagName("faultstring");
                if( list.getLength() > 0 ) {
                    Node node = list.item(0);
                    raiseError( node.getFirstChild().getNodeValue() );
                } else {
                    raiseError( "Malformed response" );
                }
            }
        }
        
        out.close();
        socket.close();
        
        return response;
    }
    
}
