package it.bigatti.iSafe;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;

import gokul.java.io.base64.*;

class Service {
    
    public static final int port = 8090;
    public static final String PREFIX = "ISAFE_DATA: ";
    
    SimpleDateFormat fmt = new SimpleDateFormat("yyyy-MM-dd");
    File storagePath = new File( System.getProperty("user.home") + File.separator + ".isafe" );

    DocumentBuilderFactory factory = null;
    DocumentBuilder builder = null;

    ServerSocket socket;
    boolean exit = false;
    
    public Service() {
        try {
            factory = DocumentBuilderFactory.newInstance();
            factory.setNamespaceAware(true);
            builder = factory.newDocumentBuilder();
        } catch( ParserConfigurationException ex ) {
            ex.printStackTrace();
            System.exit(1);
        }
    }
    
    public void accept() throws IOException {
        socket = new ServerSocket( port );
        
        while( !exit ) {
            Socket connection = socket.accept();
            InputStream in = connection.getInputStream();
            OutputStream out = connection.getOutputStream();
            
            serveRequest( in, out );
            
            in.close();
            out.close();
            connection.close();
        }
    }
    
    void serveRequest( InputStream in, OutputStream out ) throws IOException {
        DataOutputStream dout = new DataOutputStream( out );
        BufferedReader inb = new BufferedReader( new InputStreamReader( in ) );
        String line = inb.readLine();
        System.out.println( ">>"+line );
        
        //Valid iSafe data buffer
        if( line.startsWith(PREFIX)) {
            
            String contentLength = line.substring( PREFIX.length() );
            
            try {
                int length = Integer.parseInt( contentLength );
                StringBuffer sb = new StringBuffer();
                
                char[] buffer = new char[4096];
                int check = 0;
                while( check < length ) { 
                    int len = inb.read( buffer );
                    
                    sb.append( new String( buffer, 0, len ) );
                    
                    check += len;
                }
                
                dispatchRequest( sb.toString(), dout );
                
            } catch( Exception ex ) {
                ex.printStackTrace();
                send( createError( ex ), dout );
            }
        } else {
            System.out.println( "Not a valid iSafe buffer" );
        }
        
    }
    
    String createError( Exception ex ) {
        return createError( ex.getMessage() );
    }
    
    String createError( String message ) {
        String xml = "<?xml version=\"1.0\"?>"+
            "<SOAP-ENV:Envelope xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\">"+
            "  <SOAP-ENV:Header/>"+
            "  <SOAP-ENV:Fault>"+
            "    <faultcode>Server</faultcode>" +
            "    <faultstring>" + message + "</faultstring>" +
            "  </SOAP-ENV:Fault>"+
            "</SOAP-ENV:Envelope>\n";
        return xml;
    }
    
    void dispatchRequest( String message, DataOutputStream out ) throws IOException, SAXException {
        System.out.println( ">>>>"+message );
        
        Document doc = builder.parse( new ByteArrayInputStream( message.getBytes( ) ) );
        //dumpNode( doc );
        
        NodeList list = doc.getElementsByTagNameNS("http://schemas.xmlsoap.org/soap/envelope/","Body");
        if( list.getLength() > 0 ) {
            Node body = list.item(0);
            list = body.getChildNodes();
            if( list.getLength() > 1 ) {
                Node child = list.item(1);
                if( child.getNodeName().equals("getStoredData") ) {
                    send( getStoredData(), out );
                } else if ( child.getNodeName().equals("addFile") ) {
                    send( addFile( doc ), out );
                } else {
                    send( createError("Operation not recognized ("+child.getNodeName()+")"), out );
                }
            } else {
                send( createError("Unspecified request"), out );
            }
        } else {
            send( createError("Malformed request"), out );
        }
        
    }
    
    void dumpNode( Node n ) {
        dumpNode(n,0);
    }
    
    void dumpNode( Node n, int k ) {
        System.out.print("|");
        for( int j=0; j<k; j++ ) {
            System.out.print(">");
        }
        System.out.println(n.getNodeName() + "," + n.getNamespaceURI() + "," + n.getLocalName() );
        
        k++;
        NodeList list = n.getChildNodes();
        for( int i=0; i<list.getLength(); i++ ) {
            dumpNode( list.item(i), k );
        }
    }
    
    void send( String message, DataOutputStream out ) throws IOException {
        //System.out.println( "S->C: " + message );
        
        out.writeBytes( message );
    }
    
    String getStoredData() {
        String xml1 = "<?xml version=\"1.0\"?>"+
            "<SOAP-ENV:Envelope xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\">"+
            "  <SOAP-ENV:Header/>"+
            "  <SOAP-ENV:Body>"+
            "    <getStoredDataResponse>";
            
        String xml2 = "    </getStoredDataResponse>"+
            "  </SOAP-ENV:Body>"+
            "</SOAP-ENV:Envelope>\n";
            
        if( !storagePath.isFile() && !storagePath.isDirectory() ) {
            storagePath.mkdirs();
        }
        
        String[] storedFiles = storagePath.list();
        for( int i=0; i<storedFiles.length; i++ ) {
            File file = new File( storagePath + File.separator + storedFiles[i] );
            xml1 += "      <file name=\"" + storedFiles[i] + "\" date=\"" + 
                formatDate( file.lastModified() ) + 
                "\" size=\"" + file.length() + "\"/>";
        }
        
        return xml1 + xml2;
    }
    
    String addFile( Document doc ) throws IOException, SAXException {
        NodeList list = doc.getElementsByTagName("file");
        if( list.getLength() > 0 ) {
            Node node = list.item(0);
            NamedNodeMap attributes = node.getAttributes();
            
            String filename = attributes.getNamedItem("name").getNodeValue();
            FileOutputStream out = new FileOutputStream( storagePath + File.separator + filename );
            
            Node value = node.getFirstChild();
            String content = value.getNodeValue();
            
            ByteArrayInputStream bas = new ByteArrayInputStream( content.getBytes() );
            Base64InputStream in = new Base64InputStream( bas );
            
            byte[] buffer = new byte[ 4096 ];
            
            while( true ) {
                int len = in.read( buffer );
                
                if( len == -1 ) {
                    break;
                }
                
                out.write( buffer, 0, len );
            }
            
            out.close();
            bas.close();
        }
        
        return "";
    }
    
    String formatDate( long time ) {
        Date date = new Date( time );
        return fmt.format( date );
    }
}
